/*************************************************************************

This program reads the Thompson MO5 tapes into .k7 files from /dev/dsp

**************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <linux/soundcard.h>

static int fd;	/* Descriptor del stream de las muestras de audio */
static int canal=0;	/* Canal de entrada de audio 0=L, 1=R */

/*************************************************************************

1 bit reading.

Input format: Differential Manchester.

First we saturate the audio stream, so we have a train of rectangular pulses.
Then we take the absolute value of a square cycle matching convolution that 
involves 16 samples in order to make the decoding robust to noise. The result
is a typical FM signal. The rest of decoding decides what pulses are data 
and what clock.

The function rdbit() exits when a clock pulse is detected, returning the data
bit in LSB, but some static variables remains holding the neccesary data to 
continue the decoding of next bit.

**************************************************************************/

static inline char rdbit()
{
	static unsigned char sample[16];
	static int ps=0;
	static int fm=0,fm1=0,count=0;
	unsigned short ss;
	signed char a,out;
	int i;
	
	out=0;
	for(;;) {
		read(fd,&ss,2);
		if (!canal) ss>>=8; else ss&=0xff;
		a=ss;
	
		a>>=7;	/* Arithmetic shift: bit 7 is copied into all LSB's,
			   this saturate the input data */

		sample[ps]=a;	/* Data is stored in a cyclid queue */
		ps=(ps+1)&15;
	
		/* Manchester -> MFM */
		
		for (i=fm=0;i<8;i++) fm+=sample[(ps+i)&15];
		for (;i<16;i++)	     fm-=sample[(ps+i)&15];
		if (fm<0) fm=-fm;    /* absolute value */

		/* printf("%d %d\n",ss,fm); */
	
		if (fm>=1000 && fm1<1000) { /* rising edge detect */
			if (count>26) {		/* CLK pulse */
				count=0;
				fm1=fm;
				return out;
			} else if (count>9) {	/* Data ("1") pulse */
				out=1;
			} /* ...else this is a spurious pulse, so do nothing */
		}
	
		fm1=fm;
		count++;
	}
}

/*************************************************************************

This function waits for the block syncro mark.

**************************************************************************/

void sincro()
{
	unsigned long d;
	
	do {
		d<<=1;
		d|=rdbit();
	} while (d!=0x01013c5a);
}

/*************************************************************************

Simple 8 bit reading. The MSB comes first.

**************************************************************************/

unsigned char rdbyte()
{
	unsigned char d;
	int i;
	
	for (i=d=0;i<8;i++) {
		d<<=1; d|=rdbit();
	}
	return d;
}

/*************************************************************************

Whole block read. This function reads the data into the buffer pointed by the
argument and returns the type of block.

The block structure is:

	1 byte	TYPE	 (0 - Header, 1 - Data, 0xFF - End of file)
	1 byte	LEN	 (up to 256 bytes, including itself and checksum.
			  The length of 256 is coded as 0)
	N bytes	DATA	 (up to 254 bytes)
	1 byte	CHECKSUM (The addition of all the data bytes and checksum must
			  be 0)

**************************************************************************/

unsigned char rdblock(char *buf)
{
	int i,tam;
	unsigned char ck,tipo;
	sincro();
	tipo=rdbyte();
	tam=ck=rdbyte();
	if (tam==0) tam=256;
	/* fprintf(stderr,"Sincro OK Tipo :0x%02X  Tamao: 0x%02X \n",tipo,tam); */
	fflush(stderr);
	buf[0]=ck;
	for (i=1,ck=0;i<tam;i++) {
		buf[i]=rdbyte();
		ck+=buf[i];
	}
	if (ck) fprintf(stderr,"\nERROR: Checksum: 0x%02X\n",ck);
	fflush(stderr);
	return(tipo);
}


/**************************************************************************

Audio Init:  /dev/dsp @ 44100 Hz, 8 bits, stereo

***************************************************************************/

void initdsp ()
{

	int rate=44100,channels=2,bits=8;
	
	if ((fd=open("/dev/dsp", O_RDONLY)) == -1) {
		fprintf(stderr, "Cannot open sound device");
		exit(0);
	}
	if (ioctl(fd, SOUND_PCM_WRITE_RATE, &rate) == -1) {
	    fprintf(stderr, "Cannot set rate %d Hz\n",rate);
	}
	if (ioctl(fd, SOUND_PCM_WRITE_CHANNELS, &channels) == -1) {
	    fprintf(stderr, "Cannot set %ld channel(s)\n",channels);
	}
	if (ioctl(fd, SOUND_PCM_WRITE_BITS, &bits) == -1) {
	    fprintf(stderr, "Cannot set %d bits/sample\n",bits);
	}
}

/*************************************************************************

The main program reads blocks from tape and writes them to stdout.

The only exit is SIGINTR (^C) or kill

**************************************************************************/

#define FDO 1

main (int argc, char **argv)
{
	int i,tam;
	unsigned char buf[256],tipo;

	if (argc<2) {
		printf("Usage: mo5_1_rdk7 (L/R) >file.k7\n");
		exit(0);
	}

	if (*argv[1]=='r' || *argv[1]=='R') canal=1; else canal=0;

	initdsp();
	
	for(;;) {
		
		tipo=rdblock(buf);
		tam=buf[0]; if (!tam) tam=256;
		write(FDO,&tipo,1);
		write(FDO,buf,tam); 
		if (tipo==0) {
			buf[12]=0;
			fprintf(stderr,"Found: %s\n",buf+1);
		}
		fprintf(stderr,"Tipo=0x%02X Tamao=0x%02X\n",tipo,tam);
	}
}
