/*************************************************************************

Este programa lee ficheros binarios de las cintas del Thompson MO5 a partir
de /dev/dsp

**************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <linux/soundcard.h>

static int fd;	/* Descriptor del stream de las muestras de audio */

/*************************************************************************

Lectura de una muestra de sonido. Puesto que las cintas del MO5 tienen 2
pistas se impone una lectura estereo para seleccionar la pista adecuada.
En este caso ha resultado ser el canal izquierdo (L). Las muestras que se
retornan estn convertidas a 8 bits con signo, pudiendo considerarse el bit
de signo (bit 7) como el estado de la lnea.

**************************************************************************/

static inline char rdsample()
{
	short int i;
	signed char r,l;
	if (read(fd,&i,2)==0) {
		close(fd);
		exit(0);
	}
	r=i&0xff;
	l=i>>8;
	l-=128;
	return l;
}

/*************************************************************************

Decodificacin de los bits del cassette en crudo.

El formato utilizado es MANCHESTER y los datos que se modulan NRZI.

Cada bit se codifica como 416 uS en su valor y 416 uS en su complemento
Esto nos da una velocidad de 1200 baudios.

Para recuperar los bits se utiliza el siguiente algoritmo:

1 - Esperamos un cambio en el nivel lgico de los datos.
2 - Esperamos 625 uS (3/4 del tiempo de bit, o 27 muestras a 44.1 KHz).
3 - Leemos el bit.

**************************************************************************/

static inline char rdrbit()
{
	int i;
	char a;
	a=rdsample();
	while (((a^rdsample())&0x80)==0);
	for (i=0;i<27;i++) a=rdsample();	
	return a>>7;
}

/*************************************************************************

Los bits crudos obtenidos por rdrbit estn en cdigo NRZI y con cualquier
polaridad. Para recuperar los bits originales hacemos una XNOR lgica de 
cada bit con su anterior. De este modo un cambio de nivel se traduce por 
un CERO lgico mientras que el mantenimiento en el mismo nivel ser un
UNO lgico.

**************************************************************************/

static inline char rdbit()
{
	static char old=0;
	char a,b;
	a=rdrbit();
	b=a^old;
	b^=1; b&=1;
	old=a;
	return b;
}

/*************************************************************************

Esta funcin espera por la marca que nos permitir alinear los bits de 8
en 8.

**************************************************************************/

void sincro()
{
	unsigned long d;
	
	do {
		d<<=1;
		d|=rdbit();
	} while (d!=0x01013c5a);
}

/*************************************************************************

Funcin simple para la lectura de 8 bits. Ntese que el bit MSB llega en
primer lugar.

**************************************************************************/

unsigned char rdbyte()
{
	unsigned char d;
	int i;
	
	for (i=d=0;i<8;i++) {
		d<<=1; d|=rdbit();
	}
	return d;
}

/*************************************************************************

Funcin para la lectura de un bloque de datos a un buffer de memoria.
La estructura del bloque es:

	1 byte	TIPO	(0 - cabecera, 1 - datos, 0xFF - Final de fichero)
	1 byte	TAMAO	(hasta 256 bytes, incluido TAMAO y CHECKSUM.
			El tamao 256 se codifica como 0x00)
	X bytes DATOS	(hasta 254)
	1 byte	CHECKSUM (La suma de todos los bytes de DATOS y el
			CHECKSUM debe dar 0)

**************************************************************************/

unsigned char rdblock(char *buf)
{
	int i,tam;
	unsigned char ck,tipo;
	sincro();
	tipo=rdbyte();
	tam=ck=rdbyte();
	if (tam==0) tam=256;
	//fprintf(stderr,"Sincro OK Tipo :0x%02X  Tamao: 0x%02X \n",tipo,tam);
	fflush(stderr);
	buf[0]=ck;
	for (i=1,ck=0;i<tam;i++) {
		buf[i]=rdbyte();
		ck+=buf[i];
	}
	if (ck) fprintf(stderr,"\nERROR: Checksum: 0x%02X\n",ck);
	fflush(stderr);
	return(tipo);
}


/**************************************************************************

Inicializacin de /dev/dsp a 44100 Hz, 8 bits, estreo

***************************************************************************/

void initdsp ()
{

	int rate=44100,channels=2,bits=8;
	
	if ((fd=open("/dev/dsp", O_RDONLY)) == -1) {
		fprintf(stderr, "Cannot open sound device");
		exit(0);
	}
	if (ioctl(fd, SOUND_PCM_WRITE_RATE, &rate) == -1) {
	    fprintf(stderr, "Cannot set rate %d Hz\n",rate);
	}
	if (ioctl(fd, SOUND_PCM_WRITE_CHANNELS, &channels) == -1) {
	    fprintf(stderr, "Cannot set %ld channel(s)\n",channels);
	}
	if (ioctl(fd, SOUND_PCM_WRITE_BITS, &bits) == -1) {
	    fprintf(stderr, "Cannot set %d bits/sample\n",bits);
	}
}

/*************************************************************************

Lectura de un fichero binario del MO5 a stdout.

Los datos de los ficheros binarios estn precedidos por informacin 
acerca del tamao del fichero y su direccin de carga:

	1 byte	TIPO	(0 - Binario, 0xFF - BASIC, Otro - ASCII )
	1 word	TAMAO	(Tamao del fichero binario, no del bloque)
	1 word	INICIO	(Direccin de la memoria donde se ubica el fichero)
	X bytes DATOS

Los datos pueden ocupar mas de un bloque en la cinta.

**************************************************************************/

main (int argc, char **argv)
{
	int i,fdo;
	unsigned char buf[256],fn[14];
	int ini,tam;
	
	initdsp();
	
	do {
		while (rdblock(buf)!=0);	/* Primero esperamos cabecera */

		strncpy(fn,buf+1,8);
		for (i=0;i<8;i++) if (fn[i]==' ') break;
		fn[i]='.';
		strncpy(fn+i+1,buf+9,3);
		fn[i+4]=0;
		fprintf(stderr,"File: %s\n",fn);
		if (argc<2) break;
	} while (strcmp(argv[1],fn)!=0);

	fprintf(stderr,"LOADING File: %s\n",fn);

	if ((fdo=creat(fn,0644))==-1) exit(0);
	/* Ahora leemos el TAMAO del fichero y los primeros datos */

	if (rdblock(buf)!=1) fprintf(stderr,"Error de TIPO\n");
	
	if (buf[1]==0 || buf[1]==0xff) {	
		tam=buf[2]; tam<<=8; tam|=buf[3];
		ini=buf[4]; ini<<=8; ini|=buf[5];
	
		fprintf(stderr,"Prebyte: 0x%02X Inicio: 0x%04X Tamao: 0x%04X\n",buf[1],ini,tam);
		i=buf[0]; if (!i) i=256; i-=7;
		write(fdo,buf+6,i);
		tam-=i;
	} else {
		i=buf[0]; if (!i) i=256; i-=2;
		write(fdo,buf+1,i);
		tam=2000000000;
	}
	
	fprintf(stderr,"."); fflush(stderr);

	/* Seguimos leyendo bloques hasta completar el tamao del fichero
	si era binario o hasta llegar a un bloque 0xff */

	while (rdblock(buf)==1) {
		i=buf[0]; if (!i) i=256; i-=2;
		if (tam<i) i=tam;
		write(fdo,buf+1,i); 
		tam-=i;
		fprintf(stderr,"."); fflush(stderr);
	}

	close(fdo);	
	close(fd);
	fprintf(stderr,"\n\n");
}
